import time
import os
from selenium import webdriver
from selenium.webdriver.chrome.options import Options

class VideoTransmitter:
    """
    Class responsible for managing the Browser and Video Stream, from client side
    """
    def __init__(self, driver: webdriver.Chrome, server):
        """
        :param server: Chrome WebDriver Object
        """
        self.credentials = None
        self.server = server
        self.driver = driver
        
    def setCredentials(self,credentials):
        self.credentials = credentials

    def checkBrowserAlive(self):
        try:
            self.driver.current_url
            return True
        except:
            return False

    def openBrowserAndHitLink(self):
        self.driver.get(self.server)
        time.sleep(1.0)
        if(self.driver.title == "Privacy error"):
            self.driver.find_element_by_id('details-button').click()
            self.driver.find_element_by_id('proceed-link').click()
        # time.sleep(0.3)
        self.driver.find_element_by_name('username').send_keys(self.credentials.username)
        self.driver.find_element_by_name('password').send_keys(self.credentials.password)
        self.driver.find_element_by_id('submit').click()

    def close(self):
        self.driver.quit()

def createVideoTransmitter(audio=True, video=True, serverPath=None):
    options = Options()
    # options.add_argument("--headless")
    # options.add_argument('--disable-gpu')
    # options.add_argument('--window-size=1024x768')
    # options.add_argument("--no-sandbox")
    # options.add_argument("--ignore-certificate-errors")
    # options.add_argument("--remote-debugging-port=9222")

    prefs = {
        "profile.default_content_setting_values.media_stream_camera": 2,  # 1:allow, 2:block
        "profile.default_content_setting_values.media_stream_mic": 2
    }
    if audio:
        prefs["profile.default_content_setting_values.media_stream_mic"] = 1
    if video:
        prefs["profile.default_content_setting_values.media_stream_camera"] = 1

    options.add_experimental_option("prefs", prefs)
    driver = webdriver.Chrome(chrome_options=options)
    driver.maximize_window()
    return VideoTransmitter(driver,server=serverPath)

if __name__ == "__main__":
    os.environ['RCONTROLNET'] = "../../config/iorConfigs.config"

    transmitter = createVideoTransmitter()
    transmitter.openBrowserAndHitLink()
    input()
    transmitter.close()


import socket
import threading
import time
import json
import logging

class ProxyServer(threading.Thread):
    def __init__(self, server, callback):
        threading.Thread.__init__(self)
        self.server = server[0]
        self.port = server[1]
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.bind((self.server, self.port))
        self.callback = callback
        self.client = None

    def run(self) -> None:
        self.socket.listen(1)
        print("Listening for Clients")
        while True:
            socket_, addr = self.socket.accept()
            print("Client Connected ", addr)
            self.client = ProxyClient()
            self.client.connect(clientSocket=socket_, callback=self.callback)
            self.client.start()
            self.client.join()


class ProxyClient(threading.Thread):
    def __init__(self, ):
        threading.Thread.__init__(self)
        self.server = self.port = None
        self.file = self.callback = self.socket = None

    def connect(self, callback, clientSocket=None, server=None):
        if clientSocket is None:
            self.server = server[0]
            self.port = server[1]
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.connect((self.server, self.port))
        else:
            self.socket = clientSocket
        self.file = self.socket.makefile('rb')
        self.callback = callback

    def sendData(self, msg):
        self.socket.sendall(msg + b'\r\n')

    def run(self) -> None:
        while True:
            data = self.file.readline()
            data = json.loads(data.decode()[:-2])
            try:
                self.callback(data)
            except Exception as ex:
                logging.info("Exception occured while trying to callback on ProxyClient", ex)



if __name__ == "__main__":
    server = ProxyServer(server=('', 54753), callback=lambda x: print(x))
    server.start()
    server.join()
    
    time.sleep(0.5)

    client = ProxyClient()
    client.connect(server=("localhost", 54753), callback=lambda x: print(x))
    client.start()

    while True:
        client.sendData(b'Hello World')
        time.sleep(1)



import json

class PlainSerializer:
    def serialize(self, data):
        return data

class PlainDeserializer:
    def deserialize(self, data):
        return data

class JSONSerializer(PlainSerializer):
    def serialize(self, data):
        return json.dumps(data)

class JSONDeserializer(PlainDeserializer):
    def deserialize(self, data):
        return json.loads(data)


# from ior_research import IOTClient
# import time
#
# class Receiver(IOTClient):
#     def __init__(self,code,token,to,debug=False,server='iorcloud.ml'):
#         IOTClient.__init__(self,code,token,to,isTunneled=True,debug=debug,server=server)
#         assert self.connected, "Could not connect to %s"%server
#         self.transmitter_connected = False
#         self.onRequest = None
#
#     def setOnRequest(self,function):
#         if self.onRequest is not None:
#             self.onRequest.stop()
#
#         self.onRequest = RepeatedTimer(3,function)
#         self.onRequest.start()
#
#     def waitForTransmitter(self):
#         while not self.transmitter_connected:
#             if self.file.readable():
#                 data = self.readData()
#                 if data['message'] == "102":
#                     self.transmitter_connected = True
#                     self._writeline("Transmitter Connected")
#             time.sleep(0.5)
#
#     def run(self):
#         assert self.connected, "Server not connected"
#         while True:
#             if self.file.readable:
#                 data = self.readData()
#                 if data["message"] == "250":
#                     self._writeline("Connection to the other side is broken")
#                     self.transmitter_connected = False
#                     break;
#
#         if self.onRequest is not None:
#             self.onRequest.stop()
#
# from threading import Timer
#
# class RepeatedTimer(object):
#     def __init__(self, interval, function):
#         self._timer     = None
#         self.interval   = interval
#         self.function   = function
#
#         self.is_running = False
#         self.start()
#
#     def _run(self):
#         self.is_running = False
#         self.start()
#         self.function()
#
#     def start(self):
#         if not self.is_running:
#             self._timer = Timer(self.interval, self._run)
#             self._timer.start()
#             self.is_running = True
#
#     def stop(self):
#         self._timer.cancel()
#         self.is_running = False


import requests
from ior_research.utils import consts

class IORHttpClient:
    def __init__(self, server = None):
        if server is None:
            server = consts.TUNNEL_SERVER + "/api"
        self.server = server
        self.token = None
        self.verify = False

    def fetchToken(self, username, password):
        response = requests.post(self.server + "/refreshToken", json = {
            "username": username,
            "password": password
        }, headers ={
            "content-type": "application/json"
        }, verify= self.verify)
        self.token = response.json()['jwt']


import paho.mqtt.client as mqtt
from paho.mqtt.client import MQTTMessage
import json
from ior_research import SocketMessage

class Communicator:
    """
    Control Net Communicator for MQTT Protocol, it manages Protocol conversion from ControlNet Protocol to MQTT and vice versa
    """
    def __init__(self,  token: str, deviceCode: str):
        """
        token: unique identified string aka (api key)
        deviceCode: Any alpha numeric string
        """
        self.token = token
        self.deviceCode = deviceCode
        self.client = mqtt.Client()
        self.connected = None
        self.onReceive = None

    def connect(self, server="localhost"):
        """
        Connects to mqtt server on port 1883.
        """
        self.client.connect(server, 1883)
        self.client.on_connect = self.__onConnect
        self.client.loop_start()

    def __onConnect(self, client, userdata, flags, rc):
        """
        OnConnect method automatically subscribes to default topics for ControlNet and calls connected method if set by user
        client: MQTTClient object
        """
        #client.subscribe("$SYS/#")
        self.client.subscribe(self.token)
        self.client.subscribe("%s/%s"%(self.token,self.deviceCode))
        if self.connected is not None:
            self.connected(client, userdata, flags, rc)

    def __onReceive(self, client, userdata, msg):
        """
        OnReceive method invokes whenever a client message is received
        """
        if self.onReceive is not None:
            self.onReceive(msg)

    def setOnConnect(self, onConnect):
        """
        setOnConnect method sets a onConnect method.
        onConnect: takes a function
        """
        self.connected = onConnect

    def setOnReceive(self, onReceive):
        """
        setOnReceive method sets a onReceive method.
        onReceive: takes a function
        """
        self.onReceive = onReceive
        self.client.on_message = self.__onReceive

    def sendObject(self, obj, to=None):
        """
        publish message to MQTT Broker
        sm: SocketMessage object
        to: optional parameter, specifies a target client
        """
        payload = json.dumps(obj, cls=SocketMessage)
        self.sendMessage(sm=payload, to=to)
    def sendMessage(self,sm,to=None):
        """
        publish message to MQTT Broker
        sm: Message String
        to: optional parameter, specifies a target client
        """
        path = self.token
        if(to != None):
            path += "/" + to
        self.client.publish(path, sm)

def on_connect(client, userdata, flags, rc):
    print("Connected with result code "+str(rc))

def on_message(msg:MQTTMessage, payload: dict):
    print(msg.topic,payload)

def createClient(token,code,server="localhost"):
    client = Communicator(token,code)
    client.connect(server)
    return client

def createReverseClients(token, code1, code2, server="localhost"):
    """
    Create a pair of communicator clients given by common token and unique device codes
    token: A Alphanumeric String aka (api key)
    code1: device 1 code
    code2: device 2 code
    server: optional parameter, defines MQTT Broker ip address
    """
    client1 = Communicator(token, code1)
    client1.connect(server)

    client2 = Communicator(token, code2)
    client2.connect(server)

    return client1, client2

if __name__ == "__main__":
    import time
    code1,code2 = "1234","5678"

    # Creates a Pair of Communicator Clients
    client1,client2 = createReverseClients("default",code1,code2)

    # Set onConnect method
    client1.setOnConnect(on_connect)
    client2.setOnConnect(on_connect)

    # Set onReceive method
    client1.setOnReceive(on_message)
    client2.setOnReceive(on_message)

    # Let the clients initialise for 2 sec
    time.sleep(2)

    while True:
        # Sends message specific to client with code2
        client1.sendObject(SocketMessage("From 1"), to=code2)
        time.sleep(1)
        # Sends message specific to client with code1
        client2.sendObject(SocketMessage("From 2"), to=code1)




import importlib
from typing import List

import rcn
from cndi.env import getContextEnvironment
from rcn.utils import loadYamlAsClass
import os, time, sys
try:
    import ior_research
except ModuleNotFoundError:
    sys.path.append("../../")

from ior_research.utils.filterchains import MessageFilterChain
from ior_research.IOTClient import IOTClientWrapper
from ior_research.utils.video import VideoTransmitter, createVideoTransmitter
import logging
logger = logging.getLogger(__name__)

class Credentials:
    def __init__(self, username, password):
        self.username = username
        self.password = password

class ProjectConfig:
    def __init__(self, **kwargs):
        self.logger = logging.getLogger('.'.join([self.__class__.__module__, self.__class__.__name__]))
        self.clientCredentialsPath = kwargs['clientJson']
        self.clientCredentialsPath = tuple(map(lambda x:
                                            os.path.join(os.path.dirname(kwargs['controlnetConfig']),
                                                x) if not os.path.isabs(x) else x,
                                         self.clientCredentialsPath
                                         ))
        self.token = kwargs['token']
        self.videoConfigs = kwargs['streamer'] if 'streamer' in kwargs else None
        self.credentials = kwargs['credentials']
        self.credentials = Credentials(**self.credentials)
        self.filters = kwargs['filters'] if 'filters' in kwargs else list()

class Initializer:
    def __init__(self, configPath):
        self.projectConfig = loadConfig(configPath)
        self.filterChains = []
        self.transmitter = None
        self.clients = []
        self.loadFilters()

    def loadFilters(self):
        if self.projectConfig.filters is None:
            self.projectConfig.filters = list()
        for filterObj in self.projectConfig.filters:
            filter = filterObj.name
            module_elements = filter.split('.')
            module = importlib.import_module("."+module_elements[-2], '.'.join(module_elements[:-2]))
            if module_elements[-1] not in dir(module):
                logger.error(f"Filter class {module_elements[-1]}, Not found in module {'.'.join(module_elements[:-1])}")
                raise ImportError(f"Filter class {module_elements[-1]}, Not found in module {'.'.join(module_elements[:-1])}")
            classInstance = getattr(module, module_elements[-1])
            objInstance = classInstance(self, configuration=filterObj.configuration)
            self.filterChains.append(objInstance)
            logger.info(f"Filter Loaded: {filter}")


    def addFilter(self, filter: MessageFilterChain):
        self.filterChains.append(filter)

    def initializeVideoTransmitter(self) -> VideoTransmitter:
        if self.projectConfig.streamer is None:
            raise Exception("Streamer configs not supported")
        videoStreamerFlag = getContextEnvironment("rcn.initializer.video.enabled", defaultValue=True, castFunc=bool)
        if not videoStreamerFlag:
            raise Exception("Video Streaming not supported")

        videoTransmitter = createVideoTransmitter(**self.projectConfig.streamer)
        videoTransmitter.setCredentials(self.projectConfig.credentials)
        self.transmitter = videoTransmitter
        return videoTransmitter

    def processMessageInFilterChain(self, message):
        for filter in self.filterChains:
            output = filter.doFilter(message)
            if output is not None:
                message = output

    def initializeIOTWrapper(self, server="localhost", httpPort=5001, tcpPort=8000) -> List[IOTClientWrapper]:
        iotWrapperFlag = getContextEnvironment("rcn.initializer.iot.wrapper.enabled", defaultValue=True, castFunc=bool)
        if not iotWrapperFlag:
            raise Exception("IOT Wrapper Initialization not supported")


        clients = list()
        for clientPath in self.projectConfig.clientJson:
            path = os.path.abspath(clientPath)
            config = {
                "server": server,
                "httpPort": httpPort,
                "tcpPort": tcpPort,
                "useSSL": False,
                "file": path
            }

            def onReceive(message):
                self.processMessageInFilterChain(message)

            client = IOTClientWrapper(self.projectConfig.token, config=config)
            client.set_on_receive(onReceive)
            clients.append(client)
        self.clients = clients
        return clients

def loadConfig(config):
    if not os.path.isabs(config):
        config = os.path.abspath(config)
    data = loadYamlAsClass(config)
    data.credentials = Credentials(**data.credentials)

    # config = ProjectConfig(controlnetConfig=config, **data)
    return data

if __name__ == "__main__":
    initializer = Initializer("../../config/iorConfigsFrom.yml")
    videoTransmitter = initializer.initializeVideoTransmitter()
    videoTransmitter.openBrowserAndHitLink()
    while videoTransmitter.checkBrowserAlive():
        time.sleep(1)




# from ior_research import IOTClient
# import time
#
# class Transmitter(IOTClient):
#     def __init__(self,code,token,to,debug=False,server='iorcloud.ml'):
#         IOTClient.__init__(self,code,token,to,isTunneled=True,debug=debug,server=server)
#         assert self.connected, "Could not connect to %s"%server
#         self.receiver_connected = False
#
#     def waitForReceiver(self):
#         while not self.receiver_connected:
#             if self.file.readable():
#                 data = self.readData()
#                 if data['message'] ==  "100":
#                     self.receiver_connected = True
#                     self._writeline("Client Connected")
#             time.sleep(0.5)
#
#     def sendData(self,data):
#         assert self.receiver_connected, "Receiver is not connected"
#         self.file.write(data)
#         self.file.flush()
#
#     def run(self):
#         assert self.connected, "Server not connected"
#         while self.file.readable():
#             data = self.readData()
#             if data["message"] == "250":
#                 self._writeline("Connection to the other side is broken")
#                 self.receiver_connected = False
#                 break;
#             if data['message'] == "100":
#                 self.receiver_connected = True



import logging

from cndi.annotations import Autowired

from ior_research.utils.filterchains import MessageFilterChain
from ior_research.utils.text import socketMessageSchema


class MQTTPublisher(MessageFilterChain):
    logger = logging.getLogger(f"{MessageFilterChain.__module__}.{MessageFilterChain.__name__}")

    def initializeMqttClient(self, server, port, defaultTopic):
        from paho.mqtt.client import Client
        def on_connect(client: Client, userdata, flags, rc):
            self.logger.info("Connected with result code " + str(rc))
            client.subscribe(defaultTopic)

        def on_message(client: Client, userdata, msg):
            print(msg.topic + " " + str(msg.payload))

        @Autowired()
        def setMqttClient(client: Client):
            client.on_message = on_message
            client.on_connect = on_connect

            client.connect(server, port)
            client.loop_start()
            self.client = client

    def initialise(self):
        self.client = None
        self.server = self.getOrElseRaiseException("server")
        self.port = int(self.getOrElseRaiseException("port"))
        self.defaultTopic = self.getOrElseRaiseException("defaultTopic", "rcn.robot.controller")

        self.initializeMqttClient(self.server, self.port, self.defaultTopic)


    def doFilter(self,message):
        operatedMessage = socketMessageSchema.dumps(message)
        try:
            if self.client is not None:
                self.client.publish(self.defaultTopic, operatedMessage)
            else:
                self.logger.warning("MQTT Client not Initialised properly")
        except Exception as e:
            self.logger.error("Error occured while publishing the message" + e)
        return message


from builtins import super

from cndi.annotations import Autowired

from ior_research.utils.filterchains.mqtt import MQTTPublisher
from ior_research.utils.text import SocketMessage
from ior_research.vechiles.ev3 import EV3Tank


class VechileControllerFilterChain(MQTTPublisher):
    def initialise(self):
        super(VechileControllerFilterChain, self).initialise()
        from paho.mqtt.client import Client
        @Autowired()
        def setMqttClient(client: Client):
            self.vechile = EV3Tank(configuration={
                'mqtt': client
            })

    def doFilter(self,message: SocketMessage):
        if message.syncData is None:
            return message

        speed = int(message.syncData['speed'])
        steer = int(message.syncData['steer'])

        self.vechile.steer(steer, speed)

        return message


from paho.mqtt.client import Client

from cndi.annotations import Bean
import os

from cndi.env import getContextEnvironment
from cndi.exception import InvalidBeanDefination
from rcn import configDir

from ior_research.utils.consts.envs import RCONTROLNET_ENV, RCONTOLNET_PROFILE
from ior_research.utils.initializers import Initializer

@Bean()
def loadInitializer() -> Initializer:

    if RCONTROLNET_ENV in os.environ and os.path.exists(os.environ[RCONTROLNET_ENV]):
        return Initializer(os.environ[RCONTROLNET_ENV])
    elif RCONTOLNET_PROFILE in os.environ:
        configPath = os.path.join(configDir, os.environ[RCONTOLNET_PROFILE], "default.yml")
        return Initializer(configPath)
    else:
        active_profile = getContextEnvironment("rcn.profile.active", defaultValue=None)
        if active_profile is not None:
            configPath = os.path.join(configDir, active_profile, "default.yml")
            return Initializer(configPath)

    raise InvalidBeanDefination(message=f"Could not find Environment Variable {RCONTROLNET_ENV}, or path does not exist")

@Bean()
def getMqttClient() -> Client:
    return Client()


from ior_research.vechiles import IORVechile
from paho.mqtt.client import Client

class EV3Tank(IORVechile):
    def onConnect(self):
        pass
    def onMessage(self):
        pass

    def initialise(self):
        self.mqttClient: Client = self.configuration['mqtt']

    def move(self, speed):
        left_speed = speed
        right_speed = speed

        self.mqttClient.publish("ev3dev.motor.left.speed", str(left_speed))
        self.mqttClient.publish("ev3dev.motor.right.speed", str(right_speed))

    def steer(self, value, speed):
        value = min(100, value)
        value = max(-100, value)

        value /= 100

        if value != 0:
            left_speed = int(-value * speed)
            right_speed = int(value * speed)
        else:
            left_speed = speed
            right_speed = speed

        self.mqttClient.publish("ev3dev.motor.left.speed", str(left_speed))
        self.mqttClient.publish("ev3dev.motor.right.speed", str(right_speed))


class IORVechile(object):
    def __init__(self, configuration=None):
        self.configuration = configuration
        self.initialise()
    def initialise(self):
        pass
    def move(self, speed):
        pass
    def steer(self, value):
        pass


from marshmallow import Schema, fields, post_dump, post_load


class SocketMessage:
    def __init__(self, message, status=None, syncData=dict()):
        self.message = message
        self.status = status
        self.syncData=syncData


class SocketMessageSchema(Schema):
    message = fields.Str()
    status = fields.Str(required=False, default=None)
    syncData = fields.Dict(required=False, default=None)

    @post_load
    def parseObject(self, data, **kwargs):
        return SocketMessage(**data)

    @post_dump
    def remove_skip_values(self, data, **kwargs):
        print(data)
        return dict(map(lambda key: (key,data[key]), filter(lambda key: data[key] is not None, data)))

socketMessageSchema = SocketMessageSchema()


import enum
import json


class DroneAttributes(enum.Enum):
    GROUNDSPEED = 1
    AIRSPEED = 2
    BATTERY = 3
    HOME_LOCATION = 4
    HEADING = 5
    VELOCITY = 6
    LOCATION__GLOBAL_RELATIVE_FRAME = 7
    LAST_HEARTBEAT = 8
    ARMED = 9
    LOCATION__LOCAL_FRAME = 10

class MissionStatus(enum.Enum):
    MISSION_SYNCED=1
    MISSION_COMPLETED=2
    ON_MISSION=3
    MISSION_TERMINATED=4
    NONE=5

class DroneOperations(enum.Enum):
    SYNC_MISSION = 1
    START_MISSION = 2
    START_STREAMER = 3
    STOP_STREAMER = 4

class DroneActions(enum.Enum):
    SET_ALTITUDE = 1
    SET_HEADING = 2
    DIGITAL_WRITE = 3

class MessageStatus(enum.Enum):
    Initialized=1
    Connected=2
    Disconnected=3
    ERROR=4
    Operation=5
    SYNC=6
    COPTER_OPERATION=7
    MISSION_STATUS=8

class ControlNetConfigs:
    def __init__(self):
        self.token = None

    def toFile(self, path):
        with open(path, "w") as file:
            data = json.dumps(self.__dict__)
            file.write(data)

    def fromFile(self, path):
        with open(path, "r") as file:
            data = file.read()
            obj = json.loads(data)
            self.token = obj['token']


# -*- coding: utf-8 -*-
import base64
from Crypto.Cipher import AES
from urllib import parse
import uuid


IV = "1234567890123456"

class ControlNetAES(object):
    """
    Class responsible for encrypting and decrypting, the message
    """
    def __init__(self, key):
        """
        :param key: Key to perform AES Encryption
        """
        self.key = key
        self.mode = AES.MODE_CBC
        BS = len(key)
        self.pad = lambda s: s + (BS - len(s) % BS) * chr(BS - len(s) % BS)
        self.unpad = lambda s: s[0:-ord(s[-1:])]

    # Encryption function
    def encrypt(self, text):
        """
        Encrypts text and returns base64 encoded form of cipher text encoded using AES Encryption
        :param text: input raw string
        """
        cryptor = AES.new(self.key.encode("utf8"), self.mode, IV.encode("utf8"))
        self.ciphertext = cryptor.encrypt(bytes(self.pad(text), encoding="utf8"))
        # The strings obtained during AES encryption are not necessarily ASCII character sets. There may be problems when they are output to the terminal or saved. Base64 encoding is used
        return base64.b64encode(self.ciphertext)

    # Decryption function
    def decrypt(self, text):
        """
        Decrypts encrypted base64 text and returns raw string
        :param text: encrypted AES base 64 encoded string
        """
        decode = base64.b64decode(text)
        cryptor = AES.new(self.key.encode("utf8"), self.mode, IV.encode("utf8"))
        plain_text = cryptor.decrypt(decode)
        return self.unpad(plain_text)


if __name__ == '__main__':
    AES_SECRET_KEY = str(uuid.uuid4()).replace("-", '')[:16]
    print(AES_SECRET_KEY)
    aes_encrypt = ControlNetAES(AES_SECRET_KEY)
    my_email = "hello there is the world"
    e = aes_encrypt.encrypt(my_email)
    d = aes_encrypt.decrypt(e)
    print(my_email)
    print(e)
    print(d)



import sys, os

from cndi.annotations import Autowired, Component
from cndi.env import getContextEnvironment
from cndi.events import EventHandler

import threading
import time
import socket
import json
import os
import logging, base64
from ior_research.utils.aes import ControlNetAES
from ior_research.utils.serialization import JSONSerializer,JSONDeserializer

logging.basicConfig(format=f'%(asctime)s - %(name)s %(message)s', level=logging.INFO)

_STORE = dict(lastMessageRead=time.time(),
              eventHandler=None)

@Autowired()
def setConfigs(eventHandler: EventHandler):
    _STORE['eventHandler'] = eventHandler

@Component
class MessageWatcherThread(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)
        self.unexpectedDelayTriggerEventName = getContextEnvironment("rcn.ior.client.events.delay.exceeded.name", defaultValue="trigger.iorClient.delay.exceeded")
        self.expectedWaitTime = getContextEnvironment("rcn.ior.client.expected.delay", defaultValue=0.5, castFunc=float)

        self.logger = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")
        self.start()

    def run(self) -> None:
        while True:
            currentTime = time.time()
            lastMessageRead = _STORE['lastMessageRead']
            eventHandler:EventHandler = _STORE['eventHandler']
            if _STORE['eventHandler'] is not None:
                timeDiff = currentTime - lastMessageRead
                if timeDiff > self.expectedWaitTime:
                    eventHandler.triggerEventExplicit(self.unexpectedDelayTriggerEventName, timeDiff = timeDiff)
                    self.logger.debug(f"Message Watcher Time Difference Exceeded")

            time.sleep(self.expectedWaitTime/2)


class IOTClient(threading.Thread):
    """Class used to access IOR Server"""

    def __init__(self,code,token,server,time_delay = 3,key=None,debug=False,on_close = None,save_logs=False,onConnect=None, socketServer = None,
                 httpPort = 8080,tcpPort = 8000,isTunneled = False, useSSL=False):
        """
        :param code: Current Device code
        :param token: Subscription Key
        :param key: AES Encryption key, which is already defined in connections.json frile
        :param time_delay: Time Delay for a Heartbeat @Deprecated
        :param debug: See all the message in I/O stream on the CLI
        :param on_close: a function that has to be called when the connection is closed
        :param save_logs: Save Logs of all the messages, default value is False
        :param server: address of the server
        :param socket_server: (optional) only use when you have different address of ior-backend and socket-server
        :param httpPort: port to register a device on server, default value is 8080
        :param tcpPort: port on which TCP Sockets will communicate to, default value is 8000
        :param useSSL: (optional) specifies if client should communicate in http or https, default value is False
        """

        threading.Thread.__init__(self)
        #logging.basicConfig(format=f'%(asctime)s - {token}-{code} %(message)s', level=logging.INFO)

        self.__code = code
        self.__token = token
        self.__time_delay = time_delay
        self.__port = tcpPort
        self.__httpPort = httpPort
        self.__key = key
        self.useSSL = useSSL

        self.debug = debug
        self.__on_close = on_close
        self.__save_logs = save_logs
        self.__lock = threading.Lock()
        self.__server = server
        self.isTunneled = isTunneled
        self.connected = False
        self.closed = False
        self.__s = None
        self.serializer = JSONSerializer()
        self.deserializer = JSONDeserializer()
        self.lastMessageRead = time.time()
        self.setOnConnect(onConnect)

        if socketServer is None:
            self.__tunnelServer = self.__server
        else:
            self.__tunnelServer = socketServer
        self.logger = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")
        self.logger.info("*" * 80)
        self.logger.info("Using Beta - Version: %s" % self.version())
        self.logger.info("Server Configuration IP: %s" % (self.__server))
        self.logger.info("User Token %s" % self.__token)
        self.logger.info("From Code: %d" % (self.__code))
        self.logger.info("Time Delay(in Seconds): %d" % self.__time_delay)
        self.logger.info("Tunneling Enabled: " + str(self.isTunneled))
        self.logger.info("*" * 80)

        if not os.path.exists('./logs') and save_logs == True:
            os.mkdir('./logs')

        if(not self.reconnect()):
            raise Exception("Could not connect to Server at %s:%d-%d"%(self.__server,self.__httpPort,self.__port))
        self.setName("Reader-%s-%d"%(self.__token, self.__code))
        if not self.isTunneled:
            self.heartBeat = threading.Thread(target=IOTClient.__sendThread,args=(self,))
            self.heartBeat.setName("Heartbeat-%s-%d"%(self.__token,self.__code))
            self.heartBeat.start()



    @staticmethod
    def createRevertedClients(token,code,to,server="localhost",httpPort=8080,tcpPort=8000):
        client1 = IOTClient(token=token, debug=True, code=code, server=server, httpPort=httpPort,
                            tcpPort=tcpPort)
        client2 = IOTClient(token=token, debug=True, code=to, server=server, httpPort=httpPort,
                            tcpPort=tcpPort)

        return (client1,client2)

    def setOnConnect(self, on_connect):
        self.onConnect = on_connect
    @staticmethod
    def version():
        return "v0.3.7"

    def getSocket(self):
        if self.connected:
            return self.__s

    def reconnect(self):
        """
        Reconnects IOT Client to server
        """
        import requests
        if self.useSSL:
            r = requests.post('https://%s:%s/tunnel/subscribe?uuid=%s&from=%d' % (self.__server,self.__httpPort ,self.__token, self.__code), verify=False)
        else:
            r = requests.post(
                'http://%s:%s/tunnel/subscribe?uuid=%s&from=%d' % (self.__server,self.__httpPort, self.__token, self.__code))
        logging.debug("Response Status Code: %d" % r.status_code)
        if r.status_code == 404:
            self.logger.info("Request Failed")
            return False;
        if r.status_code == 409:
            raise Exception("Conflict while connecting, may another device is pre connected to the server")
        if r.status_code != 201:
            raise Exception("Invalid Credentials")

        self.logger.info("Request Successfully made to Server")
        s = r.content
        self.logger.info(s)

        if(self.__s is not None):
            self.__s.close()
            self.__s = None

        self.__s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.__s.connect((self.__tunnelServer, self.__port))
        self.__s.sendall(s);

        self.aes = ControlNetAES(self.__key)

        self.file = self.__s.makefile('rw')
        self.logger.info("Connected to Socket Server")

        self.connected = True
        if(self.onConnect is not None):
            self.onConnect()
        return True

    def __del__(self):
        self.close();

    def __sendThread(self):
        time.sleep(0.5)
        while True:
            try:
                self.__lock.acquire()
                self.getSocket().send(b"\n")
                self.logger.debug("Heartbeat message successfully send")
            except  AttributeError:
                if( self.closed):
                    self.logger.warning("Client already closed, Skipping update")
                    break;
            except ConnectionAbortedError as cae:
                self.connected = False
                logging.error("Connection Aborted",exc_info=True)
            finally:
                self.__lock.release()
            time.sleep(self.__time_delay)

    def set_on_receive(self,fn):
        self.on_receive = fn

    def __send(self,msg):
        """
        Sends Message to control net tunnel server
        """
        if(self.connected == False):
            logging.error("Server not connected Skipping")
            return False
        try:

            data = self.serializer.serialize(msg)
            data = self.aes.encrypt(data)
            self.__lock.acquire()
            self.__s.send(data + b'\r\n')
        except ConnectionAbortedError as cae:
            self.connected = False
            logging.error(cae)
        finally:
            self.__lock.release()


    def sendMessage(self,message,metadata = None,status = None):
        """
        Sends message to server, also constructs message to server acceptable format
        message: alpha numeric string
        metadata: optional object, that specificies additional data on transfer
        status: optional field, it specifies message type
        """
        msg = dict()
        msg["message"] = message
        if(status is not None):
            msg["status"] = status

        if metadata is not None:
            msg["syncData"] = metadata

        self.__send(msg)

    def close(self):
        """
        Closes the client and terminates the running Thread
        """
        self.connected = False
        self.closed = True

        self.__s.close()
        self.file.close()

        self.logger.info("Socket Closed")
        if self.__on_close != None:
            self.__on_close()

    def readData(self):
        """
        Read data sended by server
        """

        dataString = self.file.readline()
        self.lastMessageRead = time.time()
        _STORE['lastMessageRead'] = self.lastMessageRead

        if(dataString == ""):
            return None

        try:
            dataString = self.aes.decrypt(dataString)
            data = self.deserializer.deserialize(dataString)
            if(data['message'] == "HEARTBEAT"):
                return None
        except Exception as ex:
            self.logger.error(f"Error occured while receiving data: {ex}")
            data = dict(message = dataString)

        self.sendMessage("ack")
        return data

    def run(self):
        self.logger.info("Starting Thread")
        while not self.closed:
            if not self.connected:
                time.sleep(1)
                continue
            try:
                msg = self.readData()
                if msg is not None:
                    try:
                        self.on_receive(msg)
                    except Exception as ex:
                        self.logger.error("Error Occured while invoking Receive Function")
                        self.logger.error(ex)
            except socket.timeout:
                self.logger.info("socket timeout")
            except Exception as cae:
                self.connected = False
                self.logger.error("Error Occured!!!")
                self.logger.error(cae)
                break;
            time.sleep(0.01)
        self.logger.info("Thread Terminated")

class IOTClientWrapper(threading.Thread):
    """
    IOTClientWrapper is class, which wrapes IOTClient clients. It manages connection status to the server and handles IOTClient receiving messages
    """
    def __init__(self,token,config: dict = None,code = None):
        """
        Constructs object of IOTClientWrapper Class,
        """
        threading.Thread.__init__(self)
        self.config = {
            "token": token,
            "code": code,
        }

        self.logger = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")

        if config is not None:
            for key,value in config.items():
                self.config[key] = value

        if "file" in self.config:
            with open(self.config['file'], "r") as file:
                data = base64.b64decode(file.read()).decode()
                data = json.loads(data)
                self.logger.info(f"File Data: {data}")
                self.config['code'] = data['deviceCode']
                self.config['key'] = data['key']
            self.config.pop('file')


        self.closed = False
        self.set_on_receive(None)
        self.setOnConnect(None)
        self.client = None


    def setOnConnect(self, onConnect):
        self.onConnect = onConnect

    def set_on_receive(self,fn):
        """
        sets on receive function which is called everytime a message is received
        fn: function to be called when a message is received
        """
        self.fn = fn

    def terminate(self):
        """
        Terminates IOT Client connection to the server and closes the client
        """
        self.closed = True
        if self.client is not None:
            self.client.close()

    def sendMessage(self,**data):
        """
        Send message to server
        :param **data: a dict object, acceptable key-values are
            message: main message \n
            status: (optional) status of the message \n
            metadata: (optional) metadata of the message, if any
        """
        try:
            return self.client.sendMessage(**data)
        except Exception:
            return False

    def recreateClient(self):
        """
        Recreates IOT Client from config
        """
        client = IOTClient(**self.config, onConnect=self.onConnect)
        return client

    def run(self):
        self.client = self.recreateClient()
        while not self.closed:
            try:
                if self.client is None:
                    self.client = self.recreateClient()
                self.client.set_on_receive(self.fn)
                self.client.start()
                self.client.join()
                # while not self.client.closed:
                #     time.sleep(0.5)
                #     timeDiff = time.time() - self.client.lastMessageRead
                #     if timeDiff > 0.5:
                #         print("Timeout: ", timeDiff)
                self.client.close()
                self.logger.warning("Watcher Thread Closed")
                del self.client
                self.client = None
            except Exception:
                logging.error("Watcher Error: ",exc_info=True)


import setuptools

VERSION = "2.0.2a1"
INSTALLNAME = "ior_research"


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
     name=INSTALLNAME,
     version=f'v{VERSION}',
     author="Mayank Shinde",
     author_email="mayank31313@gmail.com",
     description="A platform to control robots and electronic device over Internet",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/mayank31313/ior-python",
     packages=setuptools.find_packages(),
    install_requires=["requests", "pyyaml", "pycryptodome", "paho-mqtt", "marshmallow",
              "pick>=1.2.0", "ruamel.yaml>=0.17.21", "cndi>=2.1.6a"],
     # packages=['ior_research'],
     keywords=['ior','iot','network_robos', 'control_net'],
     classifiers=[
         "Programming Language :: Python :: 3",
         "Operating System :: OS Independent",
         'Intended Audience :: Developers',
     ]
 )


from ior_research.utils.consts import DroneOperations
import json

from ior_research.utils.text import socketMessageSchema


class MessageFilterChain:
    def __init__(self, initializer, configuration=None):
        self.initializer = initializer
        self.configuration = configuration
        self.initialise()
    def getOrElseRaiseException(self, configurationAttribute, defaultValue=None):
        if configurationAttribute not in self.configuration:
            if defaultValue is not None:
                return defaultValue
            raise KeyError(f"Key {configurationAttribute} not in RCN Filter Configuration")
        return self.configuration[configurationAttribute]
    def initialise(self):
        pass
    def doFilter(self,message):
        pass;

class AirSimConnectorFilter(MessageFilterChain):
    def initialise(self):
        import airsim
        self.copter = airsim.MultirotorClient()
        self.copter.enableApiControl(True)
        self.copter.reset()
        print(self.configuration)

    def doFilter(self,message):
        # if "message" not in message:
        #     return message

        data = json.loads(message.message)
        self.copter.moveByVelocityAsync(data['roll'] * 5, data['pitch'] * 5, data['throttle'] * -1,duration=0.2)
        return message


class RControlNetMessageFilter(MessageFilterChain):
    def doFilter(self,message):
        message = socketMessageSchema.dump(message)
        return message